;; My Robot
;; Il mio primo robot...

;;(watch rules)
;;(watch facts)

(defglobal 
	?*DIST_MIN* = 100
	?*DIST_MAX* = 200
	?*DIST_POW2* = 150
	?*DIST_POW3* = 100
	?*STEP* = 15
	?*WALL_ZONE* = 100
)

;;Deftemplates

(deftemplate state
	(slot follow)
	(slot move)
	(slot power)
	(slot zone)
	(slot scan)
	(slot wallhit)
)

(deftemplate enemy
	(slot name)
	(slot x)
	(slot y)
	(slot distance)
	(slot velocity)
	(slot heading)
	(slot energy)
	(slot time)
	(slot bearing)
)

;;Genesis

(defrule set-state
	(task genesis)
	=>
	(store AdjustGunForRobotTurn true)
	(store AdjustRadarForRobotTurn true)
	(store AdjustRadarForGunTurn true)
	(assert	(state
				(follow no)
				(zone middle)
				(move ahead)
				(power 1)
				(scan no)
				(wallhit false)
			)
	)
	(printout t "GENESIS..." crlf)
)

;;Evaluate

(defrule new-enemy
	(task evaluate)
	(Time ?time)
	(ScannedRobotEvent
		(Name ?name)
		(Energy ?energy)
		(Distance ?distance)
		(Heading ?heading)
		(Velocity ?velocity)
		(Bearing ?bearing))
	(X ?myX)
	(Y ?myY)
	(Heading ?myHead)
	(not (enemy	(name ?name)))
	=>
	(assert
		(enemy 
			(name ?name)
			(energy ?energy)
			(x (enemyX ?myX ?distance ?bearing ?myHead))
			(y (enemyY ?myY ?distance ?bearing ?myHead))
			(bearing ?bearing)
			(distance ?distance)
			(heading ?heading)
			(velocity ?velocity)
			(time ?time)
		)
	)
	(printout t "New Enemy detected " ?name crlf)
)


(defrule refresh-enemy
	(task evaluate)
	(Time ?time)
	(ScannedRobotEvent
		(Name ?name)
		(Energy ?energy)
		(Distance ?distance)
		(Heading ?heading)
		(Velocity ?velocity)
		(Bearing ?bearing)
	)
	?enemy	<-	(enemy 
					(name ?name)
					(time ~?time)
				)
	(X ?myX)
	(Y ?myY)
	(Heading ?myHead)
	=>
	(modify ?enemy 
		(energy ?energy)
		(x (enemyX ?myX ?distance ?bearing ?myHead))
		(y (enemyY ?myY ?distance ?bearing ?myHead))
		(bearing ?bearing)
		(distance ?distance)
		(heading ?heading)
		(velocity ?velocity)
		(time ?time)
	)
	(printout t "Old enemy detected " ?name " -> Refresh" crlf)
)		

(defrule delete-death-people
	(task evaluate)
	(RobotDeathEvent
		(Name ?name)
	)
	?enemy	<-	(enemy
					(name ?name)
				)
	=>
	(retract ?enemy)
	(printout t "Burn Bastard " ?name ", Burn!!!" crlf)
)

;;Think

(defrule select-enemy-by-distance
	(task think)
	?state	<-	(state
					(follow ?follow)
				)
	(enemy
		(name ?name1&~?follow)
		(distance ?dist1)
	)
	(not	(enemy
				(name ?name2)
				(distance ?dist2&:(< ?dist2 ?dist1))
			)
	)
	=>
	(modify ?state
				(follow ?name1)
				(power (selectPower ?dist1))
				(zone (selectZone ?dist1))
	)	
)

(defrule select-zone
	(task think)
	?state	<-(state
			(follow ?follow)
		)
	(enemy
		(name ?follow)
		(distance ?dist)
	)
	=>
	(assert (zone (selectZone ?dist)))	
)

(defrule change-zone
	?z<-(zone ?zone)
	?state<-(state
		(zone ~?zone)
	)
	=>
	(retract ?z)	
	(modify ?state
		(zone ?zone)
	)
)
(defrule wall-up-far
	(task think)
	(Y ?y&:(< ?y (- 600 ?*WALL_ZONE*)))
	?state<-(state
			(wallhit up)
		)
	=>
	(modify ?state 
			(wallhit false)
	)
)
(defrule wall-left-far
	(task think)
	(X ?x&:(> ?x ?*WALL_ZONE*))
	?state<-(state
			(wallhit left)
		)
	=>
	(modify ?state 
			(wallhit false)
	)
)
(defrule wall-down-far
	(task think)
	(Y ?y&:(> ?y ?*WALL_ZONE*))
	?state<-(state
			(wallhit down)
		)
	=>
	(modify ?state 
			(wallhit false)
	)
)(defrule wall-right-far
	(task think)
	(X ?x&:(< ?x (- 800 ?*WALL_ZONE*)))
	?state<-(state
			(wallhit right)
		)
	=>
	(modify ?state 
			(wallhit false)
	)
)

;;Act

(defrule scan-all-around
	(task act)
	=>
	(store TurnRadarLeft 360)
	(printout t "Scan All Around" crlf)
)

(defrule aim-gun
	(task act)
	(GunHeading ?gunHead)
	(Time ?time)
	(X ?x)
	(Y ?y)
	(state
		(follow ?name)
		(power ?power)
	)
	(enemy
		(name ?name)
		(x ?enemyX)
		(y ?enemyY)
		(distance ?distance)
		(velocity ?enemyVel)
		(heading ?enemyHead)
		(time ?timeLast)
	)
	=>
	(bind ?dt (/ ?distance (- 20 (* ?power 3))))
	(bind ?px (predictX ?enemyX ?enemyVel ?enemyHead (+ ?dt ?time) ?timeLast))
	(bind ?py (predictY ?enemyY ?enemyVel ?enemyHead (+ ?dt ?time) ?timeLast))
	(bind ?dx (- ?px ?x))
	(bind ?dy (- ?py ?y))
	(bind ?alpha (atan2 ?dy ?dx))
	(bind ?beta  (normalize (- 90 (+ ?alpha ?gunHead))))
	(store TurnGunRight ?beta)
	(printout t "Tracking "  ?name crlf)
)

(defrule shot
	(task act)
	(state
		(follow ?name)
		(power ?pow)
	)
	(enemy
		(name ?name)
	)
	=>
	(store Fire ?pow)
)

(defrule move-middle
	(task act)
	(state
		(follow ?name)
		(wallhit false)
		(zone middle)
	)
	(enemy
		(name ?name)
		(bearing ?bear)
	)
	=>
	(store TurnRight (+ 90 ?bear))
	(printout t "Strafe Enemy" crlf)
)

(defrule move-near-ahead
	(task act)
	(state
		(follow ?name)
		(wallhit false)
		(move ahead)
		(zone near)
	)
	(enemy
		(name ?name)
		(bearing ?bear)
	)
	=>
	(store TurnRight (+ 120 ?bear))
	(printout t "Allontanati Ahead" crlf)
)

(defrule move-near-back
	(task act)
	(state
		(follow ?name)
		(move back)
		(wallhit false)
		(zone near)
	)
	(enemy
		(name ?name)
		(bearing ?bear)
	)
	=>
	(store TurnRight (+ 60 ?bear))
	(printout t "Allontanati Back" crlf)
)

(defrule move-far-ahead
	(task act)
	(state
		(follow ?name)
		(move ahead)
		(wallhit false)
		(zone far)
	)
	(enemy
		(name ?name)
		(bearing ?bear)
	)
	=>
	(store TurnRight (+ 60 ?bear))
	(printout t "Avvicinati Ahead" crlf)
)

(defrule move-far-back
	(task act)
	(state
		(follow ?name)
		(move back)
		(wallhit false)
		(zone far)
	)
	(enemy
		(name ?name)
		(bearing ?bear)
	)
	=>
	(store TurnRight (+ 120 ?bear))
	(printout t "Avvicinati Back" crlf)
)

(defrule hit-wall-ahead (declare (salience 100))
	(task act)
	(Heading ?head)
	?event<-(HitWallEvent
				(Bearing ?bear)
			)
	?state<-(state
				(move ahead)
		)
	=>
	(bind ?wall (selectWall (+ ?head ?bear)))
	(retract ?event)
	(modify ?state
				(move back)
				(wallhit ?wall)
	)
	(store TurnRight ?bear)
	(printout t "Hit Wall: Invert Ahead->Back " ?wall crlf)
)

(defrule hit-wall-back (declare (salience 100))
	(task act)
	?event<-(HitWallEvent
				(Bearing ?bear)
			)
	?state<-(state
				(move back)
			)
	(Heading ?head)
	=>
	(bind ?wall (selectWall (+ ?head ?bear)))
	(retract ?event)
	(modify ?state
				(move ahead)
				(wallhit ?wall)
	)
	(store TurnRight (- ?bear 180))
	(printout t "Hit Wall: Invert Back->Ahead " ?wall crlf)
)

(defrule move-ahead
	(task act)
	(state
		(move ahead)
	)
	=>
	(store Ahead ?*STEP*)
	(printout t "Moving Ahead" crlf)
)

(defrule move-back
	(task act)
	(state
		(move back)
	)
	=>
	(store Back ?*STEP*)
	(printout t "Moving Back" crlf)
)

;;Task switching

(defrule in-the-beginning
	(initial-fact)
	=>
	(assert (task genesis))
)

(defrule new-turn (declare (salience -100))
	(NewTurn)
	=>
	(assert (task evaluate))
	(printout t "NEW TURN" crlf)
	(printout t "EVALUATE..." crlf)
)

(defrule evaluate2think (declare (salience -100))
	?task <- (task evaluate)
	=>
	(retract ?task)
	(assert (task think))
	(printout t "THINK..." crlf)
)

(defrule think2act (declare (salience -100))
	?task <- (task think)
	=>
	(retract ?task)
	(assert (task act))
	(printout t "ACT..." crlf)
)

(defrule act2end (declare (salience -100))
	?task <- (task act)
	=>
	(retract ?task)
	(printout t "END TURN" crlf)
)

;;Utility functions

(deffunction enemyX(?myX ?distance ?bearing ?myHead)
	(bind ?alpha (+ ?myHead ?bearing))
	(+ ?myX (* ?distance (sin ?alpha)))
)

(deffunction enemyY(?myY ?distance ?bearing ?myHead)
	(bind ?alpha (+ ?myHead ?bearing))
	(+ ?myY (* ?distance (cos ?alpha)))
)

(deffunction predictX(?xOld ?vel ?dir ?timeNow ?timeLast)
	(bind ?dt (- ?timeNow ?timeLast))
	(bind ?dist (* ?dt ?vel))
	(bind ?dx (* ?dist (sin ?dir)))
	(+ ?xOld ?dx)
)

(deffunction predictY(?yOld ?vel ?dir ?timeNow ?timeLast)
	(bind ?dt (- ?timeNow ?timeLast))
	(bind ?dist (* ?dt ?vel))
	(bind ?dy (* ?dist (cos ?dir)))
	(+ ?yOld ?dy)
)

(deffunction distance(?x1 ?y1 ?x2 ?y2)
	(bind ?dx (- ?x1 ?x2))
	(bind ?dy (- ?y1 ?y2))
	(sqrt (+ (** ?dx 2) (** ?dy 2)))
)

(deffunction normalize(?alpha)
	(bind ?ret ?alpha)
	(if (< ?alpha -180) then (bind ?ret (+ ?alpha 360)))
	(if (> ?alpha 180) then (bind ?ret (- ?alpha 360))) 
	(bind ?alpha ?ret)
)

(deffunction selectPower(?dist)
	(bind ?pow 1)
	(if (< ?dist ?*DIST_POW2*) then (bind ?pow 2))
	(if (< ?dist ?*DIST_POW3*) then (bind ?pow 3))
	(bind ?ret ?pow)
)

(deffunction selectZone(?dist)
	(bind ?zone far)
	(if (< ?dist ?*DIST_MAX*) then (bind ?zone middle))
	(if (< ?dist ?*DIST_MIN*) then (bind ?zone near))
	(bind ?ret ?zone)
)

(deffunction selectWall(?alpha)
	(bind ?a (normalize ?alpha))
	(if (and (> ?a -45) (< ?a 45)) then (return up))
	(if (and (> ?a 45) (< ?a 135)) then (return right))
	(if (and (> ?a -135) (< ?a -45)) then (return left))
	(return down)
)